/*
 * GStreamer gst-template:
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * 
 * Original dvdreadsrc:
 * Copyright (C) 1999 Erik Walthinsen <omega@cse.ogi.edu>
 * Copyright (C) 2001 Billy Biggs <vektor@dumbterm.net>.
 * Copyright (C) 2006 Tim-Philipp Müller <tim centricular net>
 *
 * New dvdsrc, based on gst-template and dvdreadsrc:
 * Copyright (C) 2006 Jason Gerard DeRose <jderose@jasonderose.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-plugin
 *
 * <refsect2>
 * <title>Example launch line</title>
 * <para>
 * <programlisting>
 * gst-launch -v -m audiotestsrc ! plugin ! fakesink audit=TRUE
 * </programlisting>
 * </para>
 * </refsect2>
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include "gstdvdsrc.h"

/* Default property values */
#define PROP_AUDIT_DEFAULT FALSE
#define PROP_DEVICE_DEFAULT g_strdup ("/dev/dvd")
#define PROP_TITLE_DEFAULT 1
#define PROP_CHAPTER_DEFAULT 1
#define PROP_ANGLE_DEFAULT 1

/* Other helpful defines */
#define MINSEC(sec) sec / 60, sec % 60
#define SECTOR_MASK 0x7fffffff


GST_DEBUG_CATEGORY_STATIC (gst_dvdsrc_debug);
#define GST_CAT_DEFAULT gst_dvdsrc_debug


/* Signals */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};


/* Poperties */
enum
{
  PROP_0,
  PROP_AUDIT,
  PROP_DEVICE,
  PROP_TITLE,
  PROP_CHAPTER,
  PROP_ANGLE
};


/* Caps for the src pad */
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("video/mpeg, mpegversion=2, systemstream=(boolean)true"));

/* Custom formats */
static GstFormat title_format;
static GstFormat angle_format;
static GstFormat chapter_format;


/* Vmethods */
static void gst_dvdsrc_set_property (GObject * object, guint prop_id,
  const GValue * value, GParamSpec * pspec);
static void gst_dvdsrc_get_property (GObject * object, guint prop_id,
  GValue * value, GParamSpec * pspec);
static gboolean gst_dvdsrc_start (GstBaseSrc * basesrc);
static gboolean gst_dvdsrc_stop (GstBaseSrc * basesrc);
static gboolean gst_dvdsrc_is_seekable (GstBaseSrc * src);
static gboolean gst_dvdsrc_do_seek (GstBaseSrc * src, GstSegment * s);
static gboolean gst_dvdsrc_query (GstBaseSrc * basesrc, GstQuery * query);
static GstFlowReturn gst_dvdsrc_create (GstPushSrc * pushsrc, GstBuffer ** buf);

/* URI interface vmethods */
static void gst_dvdsrc_uri_handler_init (gpointer g_iface,
    gpointer iface_data);
static GstURIType gst_dvdsrc_uri_get_type (void);
static gchar ** gst_dvdsrc_uri_get_protocols (void);
static const gchar * gst_dvdsrc_uri_get_uri (GstURIHandler * handler);
static gboolean gst_dvdsrc_uri_set_uri (GstURIHandler * handler,
    const gchar * uri);

/* Private methods */
static gboolean dvdsrc_open_dvd(GstDvdSrc * src, gchar * device);
static void dvdsrc_close_dvd(GstDvdSrc * src);
static gboolean dvdsrc_open_title(GstDvdSrc * src, gint index);
static void dvdsrc_close_title(GstDvdSrc * src);
static gboolean dvdsrc_open_chapter(GstDvdSrc * src, gint index);
static gboolean dvdsrc_open_cell(GstDvdSrc * src, gint index);

static void dvdsrc_seekto_next_chapter (GstDvdSrc * src);
static void dvdsrc_seekto_next_cell (GstDvdSrc * src);
static gboolean dvdsrc_seekto_time (GstDvdSrc * src, GstClockTime ts);

static GstClockTime dvdsrc_time_at_sector (GstDvdSrc * src);
static gboolean dvdsrc_read_blocks (GstDvdSrc * src, gint blocks, 
  guint8 * buf);

static void dvdsrc_make_lang_event (GstDvdSrc * src);
static void dvdsrc_free_lang_event (GstDvdSrc * src);
static void dvdsrc_make_clut_event (GstDvdSrc * src);
static void dvdsrc_free_clut_event (GstDvdSrc * src);
static void dvdsrc_send_pending_events (GstDvdSrc * src);
static void dvdsrc_free_pending_events (GstDvdSrc * src);
static gboolean dvdsrc_audit_dvd (GstDvdSrc * src);

/* Helper functions */
static gint dvdtime_to_sec (dvd_time_t * time);
static gboolean is_nav_packet (const guint8 * data);


/* Boilerplate */
static void
_do_init (GType dvdsrc_type)
{
  static const GInterfaceInfo urihandler_info = {
    gst_dvdsrc_uri_handler_init,
    NULL,
    NULL
  };

  g_type_add_interface_static (dvdsrc_type, GST_TYPE_URI_HANDLER,
      &urihandler_info);
  
  title_format = gst_format_register ("title", "DVD title");
  angle_format = gst_format_register ("angle", "DVD angle");
  chapter_format = gst_format_register ("chapter", "DVD chapter");
}
GST_BOILERPLATE_FULL (GstDvdSrc, gst_dvdsrc, GstPushSrc, GST_TYPE_PUSH_SRC,
    _do_init);


/* base_init */
static void
gst_dvdsrc_base_init (gpointer gclass)
{
  static GstElementDetails element_details = {
    "DvdSrc",
    "Source/File/DVD",
    "DVD access using libdvdread (no Nav support)",
    "Jason Gerard DeRose <jderose@jasonderose.org>"
  };
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_set_details (element_class, &element_details);
}


/* class_init */
static void
gst_dvdsrc_class_init (GstDvdSrcClass * klass)
{
  GObjectClass * gobject_class;
  GstBaseSrcClass * gstbasesrc_class;
  GstPushSrcClass * gstpushsrc_class;

  gobject_class = (GObjectClass *) klass;
  gstbasesrc_class = (GstBaseSrcClass *) klass;
  gstpushsrc_class = (GstPushSrcClass *) klass;

  /* Vmethods */ 
  gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_dvdsrc_set_property);
  gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_dvdsrc_get_property); 
  gstbasesrc_class->start = GST_DEBUG_FUNCPTR (gst_dvdsrc_start);
  gstbasesrc_class->stop = GST_DEBUG_FUNCPTR (gst_dvdsrc_stop);
  gstbasesrc_class->is_seekable = GST_DEBUG_FUNCPTR (gst_dvdsrc_is_seekable);
  gstbasesrc_class->do_seek = GST_DEBUG_FUNCPTR (gst_dvdsrc_do_seek);
  gstbasesrc_class->query = GST_DEBUG_FUNCPTR (gst_dvdsrc_query);
  gstpushsrc_class->create = GST_DEBUG_FUNCPTR (gst_dvdsrc_create);

  /* Properties */
  g_object_class_install_property (gobject_class, PROP_AUDIT,
    g_param_spec_boolean ("audit", "Audit", 
      "Perform rigorous self-test?",
      PROP_AUDIT_DEFAULT, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DEVICE,
    g_param_spec_string ("device", "Device",
      "Path to DVD device or to folder with DVD video", 
      PROP_DEVICE_DEFAULT, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TITLE,
    g_param_spec_int ("title", "Title", 
      "Title to play",
      1, 999, PROP_TITLE_DEFAULT, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_CHAPTER,
    g_param_spec_int ("chapter", "Chapter", 
      "Chapter at which to start playing",
      1, 999, PROP_CHAPTER_DEFAULT, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_ANGLE,
    g_param_spec_int ("angle", "Angle", 
      "Angle to play", 
      1, 999, PROP_ANGLE_DEFAULT, G_PARAM_READWRITE)); 
}


/* init */
static void
gst_dvdsrc_init (GstDvdSrc * src, GstDvdSrcClass * gclass)
{
  gst_pad_use_fixed_caps (GST_BASE_SRC_PAD (src));
  gst_pad_set_caps (GST_BASE_SRC_PAD (src),
    gst_static_pad_template_get_caps (&src_factory));
  gst_base_src_set_format (GST_BASE_SRC(src), GST_FORMAT_TIME);

  src->prop_device = PROP_DEVICE_DEFAULT;
  src->prop_title = PROP_TITLE_DEFAULT;
  src->prop_chapter = PROP_CHAPTER_DEFAULT;
  src->prop_angle = PROP_ANGLE_DEFAULT;
  src->prop_audit = PROP_AUDIT_DEFAULT;
  
  src->sector = -1;
  
  src->last_uri = NULL;
  
  src->dvd_titles = -1;
  src->dvd_reader = NULL;
  src->dvd_vmg = NULL;
  
  src->title_i = -1;
  src->title_set = -1;
  src->title_ttn = -1;
  src->title_chapters = -1;
  src->title_angles = -1;
  src->title_seconds = -1;
  src->title_firstSector = -1;
  src->title_lastSector = -1;
  src->title_vts = NULL;
  src->title_file = NULL;
  src->title_tmap = NULL;
  src->title_ttu = NULL;
  
  src->chapter_i = -1;
  src->chapter_pgcSeconds = -1;
  src->chapter_firstCell = -1;
  src->chapter_lastCell = -1;
  src->chapter_firstSector = -1;
  src->chapter_lastSector = -1;
  src->chapter_pgc = NULL;
  
  src->cell_i = -1;
  src->cell_seconds = -1;
  src->cell_firstSector = -1;
  src->cell_lastSector = -1;
  src->cell_playback = NULL;
 
  src->pending_lang_event = NULL;
  src->pending_clut_event = NULL;

}


/* GObject set_property() vmethod */
static void
gst_dvdsrc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstDvdSrc *src = GST_DVDSRC (object);
  gboolean started;
  
  GST_OBJECT_LOCK (src);
  started = GST_OBJECT_FLAG_IS_SET (src, GST_BASE_SRC_STARTED);

  switch (prop_id) 
  {
    case PROP_AUDIT:
      src->prop_audit = g_value_get_boolean (value);
      break;
    case PROP_DEVICE:
      if (started) {
        GST_WARNING_OBJECT (src, 
          "Must be in GST_STATE_READY or lower to change device");
      }
      else {
         g_free (src->prop_device);
         src->prop_device = g_strdup (g_value_get_string (value));
      }
      break;
    case PROP_TITLE:
      src->prop_title = g_value_get_int (value);
      if (started)
        dvdsrc_open_title (src, src->prop_title - 1);
      break;
    case PROP_CHAPTER:
      src->prop_chapter = g_value_get_int (value);
      if (started)
        dvdsrc_open_chapter (src, src->prop_chapter - 1);
      break;
    case PROP_ANGLE:
      src->prop_angle = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
  GST_OBJECT_UNLOCK (src);
}


/* GObject get_property() vmethod */
static void
gst_dvdsrc_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstDvdSrc *src = GST_DVDSRC (object);

  switch (prop_id) {
    case PROP_AUDIT:
      g_value_set_boolean (value, src->prop_audit);
      break;
    case PROP_DEVICE:
      g_value_set_string (value, src->prop_device);
      break;
    case PROP_TITLE:
      g_value_set_int (value, src->prop_title);
      break;
    case PROP_CHAPTER:
      g_value_set_int (value, src->prop_chapter);
      break;
    case PROP_ANGLE:
      g_value_set_int (value, src->prop_angle);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


/* GstBaseSrc start() vmethod */
static gboolean
gst_dvdsrc_start(GstBaseSrc * basesrc)
{ 
  GstDvdSrc * src = GST_DVDSRC (basesrc);
  
  if (! dvdsrc_open_dvd (src, src->prop_device))
    return FALSE;

  if (! dvdsrc_open_title (src, src->prop_title - 1))
    return FALSE;

  return dvdsrc_open_chapter (src, src->prop_chapter - 1);
}


/* GstBaseSrc stop() vmethod */
static gboolean
gst_dvdsrc_stop(GstBaseSrc * basesrc)
{
  dvdsrc_close_dvd(GST_DVDSRC(basesrc));
  return TRUE;
}


/* GstBaseSrc is_seekable() vmethod */
static gboolean
gst_dvdsrc_is_seekable (GstBaseSrc * basesrc)
{
  return TRUE;
}


/* GstBaseSrc do_seek() vmethod */
static gboolean
gst_dvdsrc_do_seek (GstBaseSrc * basesrc, GstSegment * s)
{  
  GstDvdSrc *src = GST_DVDSRC (basesrc);

  if (s->format == GST_FORMAT_TIME)
    return dvdsrc_seekto_time (src, s->last_stop);
 
  return FALSE;
}


/* GstBaseSrc query() vmethod */
static gboolean 
gst_dvdsrc_query (GstBaseSrc * basesrc, GstQuery * query)
{
  GstDvdSrc *src = GST_DVDSRC (basesrc);
  GstFormat format;
  gint64 value;
  
  if (GST_QUERY_TYPE (query) != GST_QUERY_DURATION)
    return FALSE;
  
  gst_query_parse_duration (query, &format, NULL);
  
  if (format == GST_FORMAT_TIME) {
    value = src->title_seconds * GST_SECOND;
  }
  else if (format == title_format) {
    value = src->dvd_titles;
  }
  else if (format == chapter_format) {
    value = src->title_chapters;
  }
  else {
    return FALSE;
  }
  
  gst_query_set_duration (query, format, value);
  return TRUE;
}


/* GstPushSrc create() vmethod */
static GstFlowReturn
gst_dvdsrc_create (GstPushSrc * pushsrc, GstBuffer ** buf)
{
  GstDvdSrc *src = GST_DVDSRC (pushsrc);
  guint8 nav[DVD_VIDEO_LB_LEN];
  dsi_t dsi;
  gint size;
   
  /* If sector < 0, EOS has been reached */
  if (src->sector < 0) {
    dvdsrc_free_pending_events (src);
    return GST_FLOW_UNEXPECTED;
  }
  
  dvdsrc_send_pending_events (src);
  
  if (! dvdsrc_read_blocks (src, 1, nav))
    return GST_FLOW_ERROR;
    
  g_assert (is_nav_packet (nav));
  navRead_DSI (&dsi, &nav[DSI_START_BYTE]);
  g_assert (src->sector == dsi.dsi_gi.nv_pck_lbn);
  size = dsi.dsi_gi.vobu_ea + 1;
  *buf = gst_buffer_new_and_alloc (size * DVD_VIDEO_LB_LEN);
  
  if (! dvdsrc_read_blocks (src, size, GST_BUFFER_DATA (*buf)))
    return GST_FLOW_ERROR;
 
  /* Set a timestamp if src->sector in tmap */
  GST_BUFFER_TIMESTAMP (*buf) = dvdsrc_time_at_sector (src);
    
  /* Find out where to go next */
  if (dsi.vobu_sri.next_vobu == SRI_END_OF_CELL) {
    dvdsrc_seekto_next_cell (src);
  }
  else {
    src->sector += dsi.vobu_sri.next_vobu & SECTOR_MASK;
  }
  
  return GST_FLOW_OK;
}


/* Initialize URI interface */
static void
gst_dvdsrc_uri_handler_init (gpointer g_iface, gpointer iface_data)
{
  GstURIHandlerInterface * iface = (GstURIHandlerInterface *) g_iface;

  iface->get_type = gst_dvdsrc_uri_get_type;
  iface->get_protocols = gst_dvdsrc_uri_get_protocols;
  iface->get_uri = gst_dvdsrc_uri_get_uri;
  iface->set_uri = gst_dvdsrc_uri_set_uri;
}


/* GstURIHandler get_type() vmethod */
static GstURIType
gst_dvdsrc_uri_get_type (void)
{
  return GST_URI_SRC;
}


/* GstURIHandler get_protocols() vmethod */
static gchar **
gst_dvdsrc_uri_get_protocols (void)
{
  static gchar *protocols[] = { "dvd", NULL };
  return protocols;
}


/* GstURIHandler get_uri() vmethod */
static const gchar *
gst_dvdsrc_uri_get_uri (GstURIHandler * handler)
{
  GstDvdSrc *src = GST_DVDSRC (handler);
  GST_OBJECT_LOCK (src);

  g_free (src->last_uri);
  src->last_uri = g_strdup_printf ("dvd://%d,%d,%d", src->prop_title,
      src->prop_chapter, src->prop_angle);

  GST_OBJECT_UNLOCK (src);

  return src->last_uri;
}


/* GstURIHandler set_uri() vmethod */
static gboolean
gst_dvdsrc_uri_set_uri (GstURIHandler * handler, const gchar * uri)
{
  GstDvdSrc *src = GST_DVDSRC (handler);
  gchar *protocol;

  protocol = gst_uri_get_protocol (uri);
  if (protocol == NULL || !g_str_equal (protocol, "dvd")) {
    g_free (protocol);
    return FALSE;
  }


  /* parse out the new t/c/a and seek to them */
  {
    gchar *location = NULL;
    gchar **strs;
    gchar **strcur;
    gint pos = 0;

    location = gst_uri_get_location (uri);

    if (!location) {
      g_free (location);
      return FALSE;
    }

    GST_OBJECT_LOCK (src);

    src->prop_title = 1;
    src->prop_chapter = 1;
    src->prop_angle = 1;

    strcur = strs = g_strsplit (location, ",", 0);
    while (strcur && *strcur) {
      gint val;

      if (!sscanf (*strcur, "%d", &val))
        break;

      if (val <= 0) {
        g_warning ("Invalid value %d in URI '%s'. Must be 1 or greater",
            val, location);
        break;
      }

      switch (pos) {
        case 0:
          src->prop_title = val;
          break;
        case 1:
          src->prop_chapter = val;
          break;
        case 2:
          src->prop_angle = val;
          break;
      }

      strcur++;
      pos++;
    }

    if (pos > 0 && GST_OBJECT_FLAG_IS_SET (src, GST_BASE_SRC_STARTED)) {
      dvdsrc_open_title (src, src->prop_title - 1);
      dvdsrc_open_chapter (src, src->prop_chapter - 1);
    }

    GST_OBJECT_UNLOCK (src);

    g_strfreev (strs);
    g_free (location);
  }
  return TRUE;
}


/* Open DVD */
static gboolean
dvdsrc_open_dvd(GstDvdSrc * src, gchar * device)
{
  dvdsrc_close_dvd (src);

  src->dvd_reader = DVDOpen (device);
  if (src->dvd_reader == NULL) {
    GST_ELEMENT_ERROR (src, RESOURCE, READ, 
      ("Cannot open DVD at %s", device), 
      ("DVDOpen returned NULL, not pointer to dvd_reader_t"));
    return FALSE;
  }

  src->dvd_vmg = ifoOpen (src->dvd_reader, 0);
  if (src->dvd_vmg == NULL) {
    GST_ELEMENT_ERROR (src, RESOURCE, READ,
      ("Cannot open DVD at %s", device), 
      ("ifoOpen returned NULL, not pointer to ifo_handle_t"));
    return FALSE;
  }

  src->dvd_titles = src->dvd_vmg->tt_srpt->nr_of_srpts;

  if (src->prop_audit)
    return dvdsrc_audit_dvd (src);
  else
    return TRUE;
}


/* Close DVD */
static void
dvdsrc_close_dvd(GstDvdSrc * src)
{
  dvdsrc_close_title (src);
  
  src->dvd_titles = -1;
  
  if (src->dvd_vmg != NULL) {
    ifoClose (src->dvd_vmg);
    src->dvd_vmg = NULL;
  }

  if (src->dvd_reader != NULL) {
    DVDClose (src->dvd_reader);
    src->dvd_reader = NULL;
  }
}


/* Open Title */
static gboolean
dvdsrc_open_title(GstDvdSrc * src, gint index)
{
  gint i;
  title_info_t * info;

  g_assert (src->dvd_reader != NULL && src->dvd_vmg != NULL);

  if (index < 0 || index >= src->dvd_titles) {
    GST_ELEMENT_ERROR (src, RESOURCE, READ,
      ("DVD has no title %d", index + 1),
      ("Trying to open title at index %d, but DVD has %d titles",
        index, src->dvd_titles));
    return FALSE;
  }
  
  dvdsrc_close_title (src);
  
  info = & src->dvd_vmg->tt_srpt->title[index];
  src->title_i = index;
  src->title_set = info->title_set_nr;
  src->title_ttn = info->vts_ttn;
  src->title_chapters = info->nr_of_ptts;
  src->title_angles = info->nr_of_angles;
  
  src->title_vts = ifoOpen(src->dvd_reader, src->title_set);
  if (src->title_vts == NULL) {
    GST_ELEMENT_ERROR (src, RESOURCE, READ,
      ("Cannot open DVD title %d", index + 1),
      ("ifoOpen returned NULL, not pointer to ifo_handle_t"));
    return FALSE;
  }
 
  if (src->title_vts->vts_tmapt != NULL) {
    i = src->title_ttn - 1;
    if (i >= 0 && i < src->title_vts->vts_tmapt->nr_of_tmaps)
      src->title_tmap = & src->title_vts->vts_tmapt->tmap[i];
  }
  src->title_ttu = & src->title_vts->vts_ptt_srpt->title[src->title_ttn - 1];
  
  /* Check if this is an iteractive Title */
  if (src->title_chapters > 0 && src->title_ttu->ptt[0].pgn == 0) {
    GST_ELEMENT_ERROR (src, RESOURCE, READ,
      ("DVD title %d is interactive, which is not yet supported",
        src->title_i + 1),
      ("This GStreamer element does not support interactive titles"));
    return FALSE;
  }
  
  src->title_file = DVDOpenFile (src->dvd_reader, src->title_set,
                                 DVD_READ_TITLE_VOBS);
  if (src->title_file == NULL) {
    GST_ELEMENT_ERROR (src, RESOURCE, READ,
      ("Cannot open DVD title %d", index + 1),
      ("DVDOpenFile returned NULL, not pointer to dvd_file_t"));
    return FALSE;
  }
      
  if (! dvdsrc_open_chapter(src, src->title_chapters - 1))
    return FALSE;
  src->title_lastSector = src->chapter_lastSector;
  
  /* Open first chapter second so src->sector will be at title start */ 
  if (! dvdsrc_open_chapter(src, 0))
    return FALSE;
  src->title_firstSector = src->chapter_firstSector;
  src->title_seconds = src->chapter_pgcSeconds;
  
  dvdsrc_make_lang_event (src);
 
  return TRUE;  
}


/* Close Title */
static void
dvdsrc_close_title (GstDvdSrc * src)
{  
  dvdsrc_free_pending_events (src);

  src->title_i = -1;
  src->title_set = -1;
  src->title_ttn = -1;
  src->title_chapters = -1;
  src->title_angles = -1;
  src->title_seconds = -1;
  src->title_firstSector = -1;
  src->title_lastSector = -1;
  src->title_tmap = NULL;
  src->title_ttu = NULL;
  
  if (src->title_vts != NULL) {
    ifoClose (src->title_vts);
    src->title_vts = NULL;
  }

  if (src->title_file != NULL) {
    DVDCloseFile (src->title_file);
    src->title_file = NULL;
  }
}


/* Open Chapter */
static gboolean
dvdsrc_open_chapter (GstDvdSrc * src, gint index)
{
  gint pgcn, pgn, next_pgn;
  
  if (index < 0 || index >= src->title_chapters) {
    GST_ELEMENT_ERROR (src, RESOURCE, READ,
      ("DVD title has no chapter %d", index + 1),
      ("Trying to open chapter at index %d, but title has %d chapters",
        index, src->title_chapters));
    return FALSE;
  }
  
  /* Assert that VMG and VTS agree on number of Chapters in this Title */
  g_assert (src->title_chapters == src->title_ttu->nr_of_ptts);

  pgcn = src->title_ttu->ptt[index].pgcn;
  pgn = src->title_ttu->ptt[index].pgn;
  src->chapter_i = index;
  src->chapter_pgc = src->title_vts->vts_pgcit->pgci_srp[pgcn - 1].pgc;  
  src->chapter_firstCell = src->chapter_pgc->program_map[pgn - 1] - 1;
  
  if ((index + 1) == src->title_chapters) {
    src->chapter_lastCell = src->chapter_pgc->nr_of_cells;
  }
  else {
    next_pgn = src->title_ttu->ptt[index + 1].pgn;
    src->chapter_lastCell = src->chapter_pgc->program_map[next_pgn - 1] - 1;
  }
  src->chapter_pgcSeconds = dvdtime_to_sec(& src->chapter_pgc->playback_time);

  /* A chapter (of a non feature title) might have no cells, but this should
  probably be dealt with in a less hackish way */
  if (src->chapter_lastCell <= src->chapter_firstCell)
    return FALSE;

  if (! dvdsrc_open_cell (src, src->chapter_lastCell - 1))
    return FALSE;
  src->chapter_lastSector = src->cell_lastSector;

  /* Open first cell second so src->sector will be at chapter start */ 
  if (! dvdsrc_open_cell (src, src->chapter_firstCell))
    return FALSE;
  src->chapter_firstSector = src->cell_firstSector;
  
  dvdsrc_make_clut_event (src);
  
  return TRUE;
}


/* Open Cell */
static gboolean
dvdsrc_open_cell (GstDvdSrc * src, gint index)
{ 
  if (index < src->chapter_firstCell || index >= src->chapter_lastCell) {
    GST_ELEMENT_ERROR (src, RESOURCE, READ,
      ("DVD chapter has no cell %d", index),
      ("Trying to open cell %d, but cell must be >= %d and < %d",
        index, src->chapter_firstCell, src->chapter_lastCell));
    return FALSE;
  }
  
  src->cell_i = index;
  src->cell_playback = & src->chapter_pgc->cell_playback[index];
  src->cell_firstSector = src->cell_playback->first_sector;
  src->cell_lastSector = src->cell_playback->last_sector;
  src->cell_seconds = dvdtime_to_sec(& src->cell_playback->playback_time);
  
  src->sector = src->cell_firstSector;
  
  return TRUE;
}


/* Seek to next Chapter */
static void
dvdsrc_seekto_next_chapter (GstDvdSrc * src)
{
  if (src->chapter_i + 1 < src->title_chapters) {
    dvdsrc_open_chapter (src, src->chapter_i + 1);
  }
  else {
    src->sector = -1;
  }
}


/* Seek to next Cell */
static void 
dvdsrc_seekto_next_cell (GstDvdSrc * src)
{
  if (src->cell_i + 1 < src->chapter_lastCell) {
    dvdsrc_open_cell (src, src->cell_i + 1);
  }
  else {
    dvdsrc_seekto_next_chapter (src);
  } 
}


/* Seek to time ts */
static gboolean
dvdsrc_seekto_time (GstDvdSrc * src, GstClockTime ts)
{
  gint sec, i, sector, chapter, cell;

  if (src->title_tmap == NULL)
    return FALSE;
   
  sec = ts / GST_SECOND;
  i = sec / src->title_tmap->tmu - 1;
  
  if (i < 0) {
    sector = src->title_firstSector;
  }
  else {
    if (i >= src->title_tmap->nr_of_entries) {
      GST_WARNING_OBJECT (src, 
        "Probably trying to seek beyond end of title.");
      i = src->title_tmap->nr_of_entries - 1;
    }
   sector = src->title_tmap->map_ent[i] & SECTOR_MASK;
  }

  for (chapter = 0; chapter < src->title_chapters; chapter++) 
  {
    dvdsrc_open_chapter (src, chapter);
    for (cell = src->chapter_firstCell; cell < src->chapter_lastCell; cell++) 
    {
      dvdsrc_open_cell (src, cell);
      if (sector >= src->cell_firstSector &&
          sector <= src->cell_lastSector)
      {
        src->sector = sector;
        return TRUE;
      }
    }
  }
  return FALSE;
}


/* Return time at src->sector, if known */
static GstClockTime
dvdsrc_time_at_sector (GstDvdSrc * src)
{  
  gint i;
  
  if (src->sector == src->title_firstSector)
    return 0 * GST_SECOND;

  if (src->title_tmap == NULL) 
    return GST_CLOCK_TIME_NONE;
    
  for (i = 0; i < src->title_tmap->nr_of_entries; i++) 
  {
    if ((src->title_tmap->map_ent[i] & SECTOR_MASK) == src->sector)
      return src->title_tmap->tmu * (i + 1) * GST_SECOND;
  }
  
  return GST_CLOCK_TIME_NONE;
}


/* Read blocks at src->sector */
static gboolean
dvdsrc_read_blocks (GstDvdSrc * src, gint blocks, guint8 * buf)
{
  gint len;
  
  len = DVDReadBlocks (src->title_file, src->sector, blocks, buf);
  if (len == blocks)
    return TRUE;

  GST_ELEMENT_ERROR (src, RESOURCE, READ,
    ("Cannot read from DVD. The disc may be dirty or damaged."),
    ("Cannot read %d blocks at sector %d.", blocks, src->sector));
  return FALSE;
}


/* Is something wrong here?  Is this causing the Totem assertion failure:
** (totem:6785): CRITICAL **: totem_lang_table_parse_start_tag: assertion `strlen (*attr_values) == 3' failed
The same thing happens when using the original dvdreadsrc.
*/
/* Make pending_lang_event */
static void
dvdsrc_make_lang_event (GstDvdSrc * src)
{
  GstStructure *s;
  gchar lang_code[3] = { '\0', '\0', '\0' }, *t;
  gint i;
  
  dvdsrc_free_lang_event (src);

  s = gst_structure_new ("application/x-gst-dvd",
      "event", G_TYPE_STRING, "dvd-lang-codes", NULL);

  /* audio */
  for (i = 0; i < src->title_vts->vtsi_mat->nr_of_vts_audio_streams; i++) {
    const audio_attr_t * a;

    /* audio stream present? */
    
    if ((src->chapter_pgc->audio_control[i] & 0x8000) == 0)
      continue;

    a = & src->title_vts->vtsi_mat->vts_audio_attr[i];

    t = g_strdup_printf ("audio-%d-format", i);
    gst_structure_set (s, t, G_TYPE_INT, (int) a->audio_format, NULL);
    g_free (t);

    if (a->lang_type) {
      t = g_strdup_printf ("audio-%d-language", i);
      lang_code[0] = (a->lang_code >> 8) & 0xff;
      lang_code[1] = a->lang_code & 0xff;
      gst_structure_set (s, t, G_TYPE_STRING, lang_code, NULL);
      g_free (t);
    } else {
      lang_code[0] = '\0';
    }

    GST_INFO_OBJECT (src, "[%02d] Audio    %02d: lang='%s', format=%d",
        src->title_i + 1, i, lang_code, (gint) a->audio_format);
  }

  /* Subpicture stream */ 
  for (i = 0; i < src->title_vts->vtsi_mat->nr_of_vts_subp_streams; i++) {
    const subp_attr_t *u;

    /* subpicture stream present? */
    if ((src->chapter_pgc->subp_control[i] & 0x80000000) == 0)
      continue;

    u = & src->title_vts->vtsi_mat->vts_subp_attr[i];

    if (u->type) {
      t = g_strdup_printf ("subtitle-%d-language", i);
      lang_code[0] = (u->lang_code >> 8) & 0xff;
      lang_code[1] = u->lang_code & 0xff;
      gst_structure_set (s, t, G_TYPE_STRING, lang_code, NULL);
      g_free (t);
    } else {
      lang_code[0] = '\0';
    }

    GST_INFO_OBJECT (src, "[%02d] Subtitle %02d: lang='%s', type=%d",
        src->title_i + 1, i, lang_code, u->type);
  }

  src->pending_lang_event = 
      gst_event_new_custom (GST_EVENT_CUSTOM_DOWNSTREAM, s);
}


/* Free pending_lang_event */
static void
dvdsrc_free_lang_event (GstDvdSrc * src)
{
  if (src->pending_lang_event != NULL) {
    gst_event_unref (src->pending_lang_event);
    src->pending_lang_event = NULL;
  }
}


/* Make pending_clut_event */
static void
dvdsrc_make_clut_event (GstDvdSrc * src)
{
  guint * clut;
  GstStructure *structure;
  gchar name[16];
  gint i;
  
  dvdsrc_free_clut_event (src);
  
  clut = src->chapter_pgc->palette;

  structure = gst_structure_new ("application/x-gst-dvd",
      "event", G_TYPE_STRING, "dvd-spu-clut-change", NULL);

  /* Create a separate field for each value in the table. */
  for (i = 0; i < 16; i++) {
    g_snprintf (name, sizeof (name), "clut%02d", i);
    gst_structure_set (structure, name, G_TYPE_INT, (int) clut[i], NULL);
  }

  /* Create the DVD event and put the structure into it. */
  src->pending_clut_event = gst_event_new_custom (GST_EVENT_CUSTOM_DOWNSTREAM, 
    structure);
}


/* Free pending_clut_event */
static void
dvdsrc_free_clut_event (GstDvdSrc * src)
{
  if (src->pending_clut_event != NULL) {
    gst_event_unref (src->pending_clut_event);
    src->pending_clut_event = NULL;
  }
}


/* Send all pending events */
static void dvdsrc_send_pending_events (GstDvdSrc * src)
{
  GstPad *pad = GST_BASE_SRC_PAD (src);

  
  if (src->pending_lang_event != NULL) {
    gst_pad_push_event (pad, src->pending_lang_event);
    src->pending_lang_event = NULL;
  }
  
  if (src->pending_clut_event != NULL) {
    gst_pad_push_event (pad, src->pending_clut_event);
    src->pending_clut_event = NULL;
  }
}


/* Free all pending events */
static void dvdsrc_free_pending_events (GstDvdSrc * src)
{
  dvdsrc_free_lang_event (src);
  dvdsrc_free_clut_event (src);
}


/* Audit, self-test */
/* This is pretty messy looking, but the important thing is that it be thorough */
static gboolean
dvdsrc_audit_dvd (GstDvdSrc * src)
{
  gint i, j, k, title, chapter, sec, sector;
  
  /* The readability isn't very good using GST_INFO_*, thus the printf */
  printf ("DvdSrc Audit: performing rigorous self-test using current DVD\n");
  printf ("DVD is reporting %d titles\n", src->dvd_titles);

  for (i = 0; i < src->dvd_titles; i++)
  {
    title = i + 1;
    printf ("Testing title %d:\n", title);
    dvdsrc_open_title (src, i);


    /* Basic reporting, continue if vts_tmapt==NULL */
    printf ("  Title has duration %d:%02d\n", MINSEC(src->title_seconds));
    printf ("  Title has %d chapters\n", src->title_chapters);
    printf ("  Title has %d angles\n", src->title_angles);
    if (1 != src->title_angles)
      return FALSE;
    if (src->title_vts->vts_tmapt == NULL) {
      printf ("  Title does not looks like a feature title (no vts_tmapt)\n");
      printf ("  No usefull tests to do on this title...\n");
      continue;
    }
    printf ("  Title looks like a feature title (has vts_tmapt)\n");


    /* Check that all chapter_pgc report same title duration */
    printf ("  Checking that all chapter_pgc report same title duration\n");
    for (j = 0; j < src->title_chapters; j++) {
      chapter = j + 1;
      dvdsrc_open_chapter (src, j);
      if (src->title_seconds != src->chapter_pgcSeconds) {
        printf ("    Chapter %d %d:%02d FAIL\n", chapter, MINSEC(src->chapter_pgcSeconds));
        return FALSE;
      }
      printf ("    Chapter %d %d:%02d, %d, %d OKAY\n", chapter, MINSEC(src->chapter_pgcSeconds), src->chapter_firstSector, src->chapter_lastSector);
    }
    printf ("  All chapter_pgc report same title duration: OKAY\n");
    
    
    /* Check that sum of all cell_seconds <= title_seconds */
    /* For multi-angle DVDs, this test should show whether cells that are being
       included should not be.  The sum of the cells is typically slightly
       less than title_seconds, I assume from truncation error. */
    printf ("  Checking that sum of cell_seconds <= title_seconds\n");
    sec = 0;
    dvdsrc_open_chapter (src, 0);
    while (src->sector >= 0) {
      printf ("    Cell %d = %d:%02d\n", src->cell_i, MINSEC(src->cell_seconds));
      sec += src->cell_seconds;
      dvdsrc_seekto_next_cell (src);
    }
    printf ("  Sum of cell_seconds (%d:%02d) <= title_seconds (%d:%02d): ",
        MINSEC (sec), MINSEC (src->title_seconds));
    if (sec <= src->title_seconds) {
      printf ("OKAY\n");
    }
    else {
      printf ("FAIL\n");
      return FALSE;
    }
    
    /* Report some tmap related stuff */
    printf ("  Title %d is in title_set %d\n", title, src->title_set);
    printf ("  From nr_of_tmaps, title_set appears to have %d titles\n",
            src->title_vts->vts_tmapt->nr_of_tmaps);
    printf ("  Title should be using tmap at index %d\n", src->title_ttn - 1);
    if (src->title_tmap == NULL) {
      printf ("  Title has a tmap, index in range: FAIL\n");
      return FALSE;
    }
    printf ("  Title has a tmap, index in range: OKAY\n");
    printf ("  Tmap has %d entries at intervals of %d seconds\n",
            src->title_tmap->nr_of_entries, src->title_tmap->tmu);


    /* Test is broken for duration==0, but that's a degenerate case anyway.
       The strange thing is that I've found DVDs with a duration==0 title
       that has a tmap. */
    if (src->title_seconds <= 0) {
      printf ("  Title_seconds <= 0, not checking match between tmap and title\n");
      continue;
    }
    
    
    /* Check for time match between tmap and title */
    printf ("  Checking time match between tmap and title...\n");
    sec = src->title_tmap->tmu * src->title_tmap->nr_of_entries;
    printf ("    Title has duration %d:%02d\n", MINSEC(src->title_seconds));
    printf ("    Last tmap entry is at time %d:%02d\n", MINSEC(sec));
    if (sec > src->title_seconds) {
      printf ("    Last_tmap_entry <= title_seconds: FAIL\n");
      return FALSE;
    }
    printf ("    Last_tmap_entry <= title_seconds: OKAY\n");
    /* Some DVDs (e.g., Kill Bill Vol 2) will fail for + (1 * tmu), thus the
       (2 * tmu) */
    printf ("    Last_tmap_entry + (2 * tmu) >= title_seconds:");
    if (sec + (2 * src->title_tmap->tmu) >= src->title_seconds) {
      printf ("OKAY\n");
    }
    else {
      printf ("FAIL\n");
      return FALSE;
    }
    printf ("  Tmap and title match on time: OKAY\n");
    
    
    /* Check for sector match between tmap and title */
    printf ("  Checking sector match between tmap and title...\n");
    printf ("    First sector in title is %d\n", src->title_firstSector);
    sector = src->title_tmap->map_ent[0] & SECTOR_MASK;
    printf ("    First sector in tmap is  %d\n", sector);
    if (sector < src->title_firstSector) {
      printf ("    First_tmap_sector >= first_title_sector: FAIL\n");
      return FALSE;
    }
    printf ("    First_tmap_sector >= first_title_sector: OKAY\n");
    printf ("    Last sector in title is  %d\n", src->title_lastSector);
    sector = src->title_tmap->map_ent[src->title_tmap->nr_of_entries - 1] & SECTOR_MASK;
    printf ("    Last sector in tmap is   %d\n", sector);
    if (sector > src->title_lastSector) {
      printf ("    Last_tmap_sector <= last_title_sector: FAIL\n");
      return FALSE;
    }
    printf ("    Last_tmap_sector <= last_title_sector: OKAY\n");
    printf ("  Tmap and title match on sector: OKAY\n");
    
    printf ("Finished testing title %d: OKAY\n", title);
  }
  printf ("DvdSrc Audit: OKAY\n");
  return TRUE;
}

 
/* Convert a dvd_time_t struct to seconds */
static gint
dvdtime_to_sec (dvd_time_t * time)
{
  gint sec;
  const gint hour = 3600;
  const gint min = 60;

  g_return_val_if_fail ((time->hour >> 4) < 0xa
      && (time->hour & 0xf) < 0xa, -1);
  g_return_val_if_fail ((time->minute >> 4) < 0x7
      && (time->minute & 0xf) < 0xa, -1);
  g_return_val_if_fail ((time->second >> 4) < 0x7
      && (time->second & 0xf) < 0xa, -1);

  sec = hour * ((time->hour >> 4) * 10 + (time->hour & 0xf));
  sec += min * ((time->minute >> 4) * 10 + (time->minute & 0xf));
  sec += ((time->second >> 4) * 10 + (time->second & 0xf));

  return sec;
}


/* Returns true if the pack is a NAV pack */
static gboolean
is_nav_packet (const guint8 * data)
{
  if (GST_READ_UINT32_BE (data + 0x26) != 0x000001BF)
    return FALSE;

  /* Check that this is substream 0 (PCI) */
  if (data[0x2c] != 0)
    return FALSE;

  if (GST_READ_UINT32_BE (data + 0x400) != 0x000001BF)
    return FALSE;

  /* Check that this is substream 1 (DSI) */
  if (data[0x406] != 1)
    return FALSE;

  /* Check sizes of PCI and DSI packets */
  if (GST_READ_UINT16_BE (data + 0x2a) != 0x03d4)
    return FALSE;

  if (GST_READ_UINT16_BE (data + 0x404) != 0x03fa)
    return FALSE;

  return TRUE;
}


/* Initialize plugin */
static gboolean
plugin_init (GstPlugin * plugin)
{
  /* exchange the strings 'plugin' and 'Template plugin' with your
   * plugin name and description */
  GST_DEBUG_CATEGORY_INIT (gst_dvdsrc_debug, "dvdsrc",
      0, "DVD access using libdvdread (no Nav support)");

  return gst_element_register (plugin, "dvdsrc",
      GST_RANK_PRIMARY, GST_TYPE_DVDSRC);
}


/* Register plugin */
GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "gstdvdsrc",
    "DVD access using libdvdread (no Nav support)",
    plugin_init, VERSION, "LGPL", "GStreamer", "http://gstreamer.net/")
